% Script - MATALB package of Underwater Acoustic Localization
% Compute the RMSE (miss distance) vs. the SNR for the semi blind 
% localization (SBL), matched filed processing (MFP) and GCC-PHAT methods
%
% "Static" Scenario (stationary flat surface, flat bottom)
%
% Other m-files required: compute_SBL_and_oracleMFP3_objective_function.m
%                         compute_GCCPHAT_cost_function_efficient.m
%                         compute_SBL_cost_function_efficient.m
%                         compute_oracleMFP3_cost_function.m
%                         compute_GCCPHAT_cost_function_efficient.m
%                         
% Subfunctions: compute_SBL_and_oracleMFP3_objective_function,
%               compute_GCCPHAT_cost_function_efficient,
%               compute_SBL_cost_function_efficient,
%               compute_oracleMFP3_cost_function,
%               compute_GCCPHAT_cost_function_efficient
% MAT-files required: none
%
% Based on the paper:
%			 "Direct Localization in Underwater Acoustics via Convolutional 
%             Neural Networks: A Data-Driven Approach", Amir Weiss,
%             Toros Arikan, and Gregory W. Wornell,
%			  in Proc. of MLSP, August 2022.
%
% See also: supplementary_materials_DLOC_MLSP2022.pdf for more information about this MATLAB package
% Author: Amir Weiss
% email: amirweiss15@gmail.com
% Website: http://www.weissamir.com
% July 2022; Last revision: 11-Jul-2022
%------------- BEGIN CODE --------------
clear;
clc;

T = 100; % smaple size
SNR_vec_dB = linspace(-10,20,10);
SNR_vec_lin = 10.^(SNR_vec_dB/10);
noise_power_vec = 1./SNR_vec_lin;
L = 4; % number of arrays
fmax = 1000; % [Hz]

fft_coef = 1/sqrt(T);
ifft_coef = sqrt(T);

x_receiver = [150 ; 50 ; -50 ; -150]; % x coordinate of the sensors [m]
y_receiver = [-250 ; -250 ; -250 ; -250]; % y coordinate of the sensors [m]
z_receiver = [10 ; 15 ; 20 ; 25]; % z coordinate of the sensors [m]
p_array = [x_receiver' ; y_receiver' ; z_receiver'];

c = 1500;
h = 50; % h is the depth of the area considered
w = (0:1:T-1)*2*pi*fmax/T;

% Coarse Grid
x_candidate = linspace(-150,150,50);
y_candidate = linspace(-100,0,50);
z_candidate = linspace(5,h-5,40);

dim_x_cand = length(x_candidate);
dim_y_cand = length(y_candidate);
dim_z_cand = length(z_candidate);

ind_vec = 1:(dim_x_cand*dim_y_cand*dim_z_cand);
length_ind_vec = dim_x_cand*dim_y_cand*dim_z_cand;

num_of_trials = 1000;

SBL_error_tensor = zeros(length(SNR_vec_lin),3,num_of_trials);
MFP3_error_tensor = zeros(length(SNR_vec_lin),3,num_of_trials);
GCCPHAT_error_tensor = zeros(length(SNR_vec_lin),3,num_of_trials);

p_source_test = zeros(length(SNR_vec_lin),num_of_trials,3);
range_source_test = zeros(length(SNR_vec_lin),num_of_trials);
DOA_source_test = zeros(length(SNR_vec_lin),num_of_trials);
phi_source_test = zeros(length(SNR_vec_lin),num_of_trials);

display('"Static" setting, no perturbation, -10 to 20 dB')

tic
h_SNR = waitbar(0,'SNR Progress');
for SNR_ind=1:length(SNR_vec_lin)
   
    noise_power = noise_power_vec(SNR_ind);
    
    h_MC = waitbar(0,'MC Progress');
    for mc_ind=1:num_of_trials
        
        x0 = 300*rand-150; % true x coordinate of emitters [m]
        y0 = 100*rand-100; % true y coordinate of emitters [m]
        z0 = 5 + rand*40; % the depth of the source
        p = [x0 ; y0 ; z0];
        p_source_test(SNR_ind,mc_ind,:) = p;
        range_source_test(SNR_ind,mc_ind) = norm(p);
        DOA_source_test(SNR_ind,mc_ind) = atan2(y0,x0) + pi;
        phi_source_test(SNR_ind,mc_ind) = atan(sqrt(y0^2+x0^2)/z0);
        
        surface_time_delay_std = (norm(p-mean(p_array,1))/c)*0;
        
        S = (randn(1,T) + 1i*randn(1,T))./sqrt(2);
        
        x_bar = zeros(T,L);
        b_oracleMFP3 = zeros(3,L);
        D_true = zeros(3,T,L);
        for l_ind=1:L
            % Option A: random attenuation coefficients
            b = 0.995+0.1*(randn(3,1) + 1i*randn(3,1))./sqrt(2); % E[|b_{rl}|^2]=1
            b_oracleMFP3(:,l_ind) = b;

            distance_source_reciver = sqrt( (x0-x_receiver(l_ind)).^2 + (y0-y_receiver(l_ind)).^2 + (z0-z_receiver(l_ind)).^2 );
            tau1 = distance_source_reciver/c;
            horizontal_distance = sqrt( (x0-x_receiver(l_ind)).^2 + (y0-y_receiver(l_ind)).^2 );
            distance_surface_bounce = sqrt ( horizontal_distance^2 + (z0 + z_receiver(l_ind))^2 );
            tau2 = distance_surface_bounce/c;
            
            flag_surface_perturbation = 0;
            while flag_surface_perturbation==0
                tau2 = distance_surface_bounce/c + surface_time_delay_std*randn;
                if tau2>0
                    flag_surface_perturbation = 1;
                end
            end
            
            distance_bottom_bounce = sqrt ( horizontal_distance^2 + (2*h - z0 - z_receiver(l_ind))^2 );
            tau3 = distance_bottom_bounce/c;
            tau = [tau1 ; tau2 ; tau3];

            V = sqrt(noise_power)*(randn(1,T) + 1i*randn(1,T))./sqrt(2);
            
            D_true(:,:,l_ind) = exp(-1i*tau*w);
            x_bar(:,l_ind) = S.*(b.'*D_true(:,:,l_ind)) + V;
        end
        C_SBL = zeros(1,dim_x_cand*dim_y_cand*dim_z_cand);
        C_MFP3 = zeros(1,dim_x_cand*dim_y_cand*dim_z_cand);
        C_GCCPHAT = zeros(1,dim_x_cand*dim_y_cand*dim_z_cand);
        
        parfor ind_parfor_loop=1:length_ind_vec
            [i_x,i_y,i_z] = ind2sub([dim_x_cand,dim_y_cand,dim_z_cand],ind_vec(ind_parfor_loop));
            x_cand = x_candidate(i_x);
            y_cand = y_candidate(i_y);
            z_cand = z_candidate(i_z);
            
            [C_SBL(ind_parfor_loop),C_MFP3(ind_parfor_loop)] = compute_SBL_and_oracleMFP3_objective_function(p_array,x_bar,h,w,[x_cand; y_cand; z_cand],c,b_oracleMFP3);
            C_GCCPHAT(ind_parfor_loop) = compute_GCCPHAT_cost_function_efficient(p_array,x_bar,w,[x_cand; y_cand; z_cand],c);
        end
        
        % SBL
        [max_val_SBL,max_ind_SBL] = max(C_SBL);
        [i_x_max_SBL,i_y_max_SBL,i_z_max_SBL] = ind2sub([dim_x_cand,dim_y_cand,dim_z_cand],max_ind_SBL);
        p_est_SBL_grid = [x_candidate(i_x_max_SBL) ; y_candidate(i_y_max_SBL) ; z_candidate(i_z_max_SBL)];
        
        ObjectiveFunction_SBL = @(p_cand) -compute_SBL_cost_function_efficient(p_array,x_bar,h,w,p_cand,c);
        p_est_SBL = fminsearch(ObjectiveFunction_SBL,p_est_SBL_grid);
        
        SBL_error_tensor(SNR_ind,:,mc_ind) = p-p_est_SBL;
        
        % MFP
        [max_val_MFP3,max_ind_MFP3] = max(C_MFP3);
        [i_x_max_MFP3,i_y_max_MFP3,i_z_max_MFP3] = ind2sub([dim_x_cand,dim_y_cand,dim_z_cand],max_ind_MFP3);
        p_est_MFP3_grid = [x_candidate(i_x_max_MFP3) ; y_candidate(i_y_max_MFP3) ; z_candidate(i_z_max_MFP3)];
        
        ObjectiveFunction_MFP3 = @(p_cand) -compute_oracleMFP3_cost_function(p_array,x_bar,h,w,p_cand,c,b_oracleMFP3);
        
        p_est_MFP3 = fminsearch(ObjectiveFunction_MFP3,p_est_MFP3_grid);
        
        MFP3_error_tensor(SNR_ind,:,mc_ind) = p-p_est_MFP3;
        
        % GCCPHAT
        [max_val_GCCPHAT,max_ind_GCCPHAT] = max(C_GCCPHAT);
        [i_x_max_GCCPHAT,i_y_max_GCCPHAT,i_z_max_GCCPHAT] = ind2sub([dim_x_cand,dim_y_cand,dim_z_cand],max_ind_GCCPHAT);
        p_est_GCCPHAT_grid = [x_candidate(i_x_max_GCCPHAT) ; y_candidate(i_y_max_GCCPHAT) ; z_candidate(i_z_max_GCCPHAT)];
        
        ObjectiveFunction_GCCPHAT = @(p_cand) -compute_GCCPHAT_cost_function_efficient(p_array,x_bar,w,p_cand,c);
        
        p_est_GCCPHAT = fminsearch(ObjectiveFunction_GCCPHAT,p_est_GCCPHAT_grid);
        
        GCCPHAT_error_tensor(SNR_ind,:,mc_ind) = p-p_est_GCCPHAT;
        
        waitbar(mc_ind/num_of_trials,h_MC);
        display(['MC loop: ',num2str(mc_ind),' out of ',num2str(num_of_trials)]);
    end
    close(h_MC)
    waitbar(SNR_ind/length(SNR_vec_lin),h_SNR);
    display(['SNR loop: ',num2str(SNR_ind),' out of ',num2str(length(SNR_vec_lin))]);
end
close(h_SNR);
total_time_of_simulation = toc;

SBL_avg_RMS_miss_distance = sqrt(mean(mean(SBL_error_tensor.^2,3),2));
MFP3_avg_RMS_miss_distance = sqrt(mean(mean(MFP3_error_tensor.^2,3),2));
GCCPHAT_avg_RMS_miss_distance = sqrt(mean(mean(GCCPHAT_error_tensor.^2,3),2));
[SBL_avg_RMS_miss_distance MFP3_avg_RMS_miss_distance GCCPHAT_avg_RMS_miss_distance]
display(['Simulation runtime: ',num2str(total_time_of_simulation),' [sec]'])

figure
plot(SNR_vec_dB,GCCPHAT_avg_RMS_miss_distance,'LineWidth',1.5);
hold;
plot(SNR_vec_dB,SBL_avg_RMS_miss_distance,'LineWidth',1.5);
plot(SNR_vec_dB,MFP3_avg_RMS_miss_distance,'LineWidth',1.5);
grid;
xlabel('SNR [dB]');
ylabel('RMSE [m]');
legend('GCC-PHAT','SBL','Oracle MFP3','Location','northeast');
%------------- END CODE --------------