# Script for testing the model for direct localization
# Based on the paper:
#			 "Direct Localization in Underwater Acoustics via Convolutional 
#             Neural Networks: A Data-Driven Approach", Amir Weiss,
#             Toros Arikan, and Gregory W. Wornell,
#			  in Proc. of MLSP, August 2022.
#
# See also: supplementary_materials_DLOC_MLSP2022.pdf for more information about this Python package
# Author: Amir Weiss
# email: amirweiss15@gmail.com
# Website: http://www.weissamir.com
# July 2022; Last revision: 11-Jul-2022
#------------- BEGIN CODE --------------
import tensorflow as tf
import tensorflow.keras as k
from tensorflow.keras import layers
from tensorflow.keras.models import Model

import numpy as np
import matplotlib.pyplot as plt
from tqdm import tqdm
import time
import pickle
import os

import random 

import scipy.io as sio

def DOA_loss_fn(y_true, y_pred):
    cyclic_error = 2 - 2*tf.math.cos( (y_pred - y_true) )
    return tf.reduce_mean(cyclic_error, axis=-1)  # Note the `axis=-1`


def Elevation_my_loss_fn(y_true, y_pred):
    cyclic_error = 2 - 2*tf.math.cos( 2*(y_pred - y_true) )
    return tf.reduce_mean(cyclic_error, axis=-1)  # Note the `axis=-1`


def direct_localization_loss_fn_spherical(y_true, y_pred):
    
    DOA_true = y_true[:,0]
    Range_true = y_true[:,1]
    Elevation_true = y_true[:,2]
    
    DOA_pred = y_pred[:,0]
    Range_pred = y_pred[:,1]
    Elevation_pred = y_pred[:,2]
    
    term_A = tf.math.sin(Elevation_true)*tf.math.sin(Elevation_pred)*tf.math.cos(DOA_true-DOA_pred)
    
    term_B = tf.math.cos(Elevation_true)*tf.math.cos(Elevation_pred)
    
    sum_terms_A_B = term_A + term_B
    
    spherical_distance_square = tf.math.square(Range_true) + tf.math.square(Range_pred) - 2*Range_true*Range_pred*sum_terms_A_B
    
    return tf.reduce_mean(spherical_distance_square, axis=-1)  # Note the `axis=-1`


def get_custom_model_range_est(in0, input_shape, T, k_sz=3, long_k_sz=101, lr=0.0003):

    start_neurons = 8

    num_of_functions = input_shape[0]

    x = in0
    x = layers.BatchNormalization()(x)

    # 200
    conv1 = layers.Conv2D(T, (num_of_functions, long_k_sz), activation="relu", padding="same")(x)
    conv1 = layers.Conv2D(T, (num_of_functions, long_k_sz), activation="relu", padding="same")(conv1)
    conv1 = layers.Conv2D(start_neurons * 8, (num_of_functions, k_sz), activation="relu", padding="same")(conv1)
    pool1 = layers.AveragePooling2D((1, 2))(conv1)
    pool1 = layers.Dropout(0.5)(pool1)

    # 100
    conv2 = layers.Conv2D(start_neurons * 8, (k_sz, k_sz), activation="relu", padding="same")(pool1)
    conv2 = layers.Conv2D(start_neurons * 8, (k_sz, k_sz), activation="relu", padding="same")(conv2)
    pool2 = layers.AveragePooling2D((1, 2))(conv2)
    pool2 = layers.Dropout(0.5)(pool2)

    # 50
    conv3 = layers.Conv2D(start_neurons * 4, (k_sz, k_sz), activation="relu", padding="same")(pool2)
    conv3 = layers.Conv2D(start_neurons * 4, (k_sz, k_sz), activation="relu", padding="same")(conv3)
    pool3 = layers.AveragePooling2D((1, 2))(conv3)
    pool3 = layers.Dropout(0.5)(pool3)


    # 25
    conv4 = layers.Conv2D(start_neurons * 2, (k_sz, k_sz), activation="relu", padding="same")(pool3)
    conv4 = layers.Conv2D(start_neurons * 2, (k_sz, k_sz), activation="relu", padding="same")(conv4)
    pool4 = layers.AveragePooling2D((1, 2))(conv4)
    pool4 = layers.Dropout(0.5)(pool4)

    # 12
    convm = layers.Conv2D(start_neurons * 8, (k_sz, k_sz), activation="relu", padding="same")(pool4)
    convm = layers.Conv2D(start_neurons * 8, (k_sz, k_sz), activation="relu", padding="same")(convm)
    poolm = layers.AveragePooling2D((1, 2))(convm)

    # 3
    two_before_output_layer = layers.Flatten()(poolm)
    one_before_output_layer = layers.Dense(10,activation='relu')(two_before_output_layer)
    output_layer = layers.Dense(1,activation='relu')(one_before_output_layer)
    x_out_range = output_layer

    supreg_net = Model(in0, x_out_range, name='supreg')

    supreg_net.compile(optimizer=k.optimizers.Adam(learning_rate=lr),
              loss=tf.keras.losses.MeanSquaredError(),
              metrics=[tf.keras.losses.MeanSquaredError()])
    
    return supreg_net, x_out_range, one_before_output_layer


def get_custom_model_DOA_est_MCE(in0, input_shape, T, k_sz=3, long_k_sz=101, lr=0.0003):
    
    start_neurons = 8

    num_of_functions = input_shape[0]

    x = in0
    x = layers.BatchNormalization()(x)

    # 200
    conv1 = layers.Conv2D(T, (num_of_functions, long_k_sz), activation="relu", padding="same")(x)
    conv1 = layers.Conv2D(T, (num_of_functions, long_k_sz), activation="relu", padding="same")(conv1)
    conv1 = layers.Conv2D(start_neurons * 8, (num_of_functions, k_sz), activation="relu", padding="same")(conv1)
    pool1 = layers.AveragePooling2D((1, 2))(conv1)
    pool1 = layers.Dropout(0.5)(pool1)

    # 100
    conv2 = layers.Conv2D(start_neurons * 8, (k_sz, k_sz), activation="relu", padding="same")(pool1)
    conv2 = layers.Conv2D(start_neurons * 8, (k_sz, k_sz), activation="relu", padding="same")(conv2)
    pool2 = layers.AveragePooling2D((1, 2))(conv2)
    pool2 = layers.Dropout(0.5)(pool2)

    # 50
    conv3 = layers.Conv2D(start_neurons * 4, (k_sz, k_sz), activation="relu", padding="same")(pool2)
    conv3 = layers.Conv2D(start_neurons * 4, (k_sz, k_sz), activation="relu", padding="same")(conv3)
    pool3 = layers.AveragePooling2D((1, 2))(conv3)
    pool3 = layers.Dropout(0.5)(pool3)


    # 25
    conv4 = layers.Conv2D(start_neurons * 2, (k_sz, k_sz), activation="relu", padding="same")(pool3)
    conv4 = layers.Conv2D(start_neurons * 2, (k_sz, k_sz), activation="relu", padding="same")(conv4)
    pool4 = layers.AveragePooling2D((1, 2))(conv4)
    pool4 = layers.Dropout(0.5)(pool4)

    # 12
    convm = layers.Conv2D(start_neurons * 8, (k_sz, k_sz), activation="relu", padding="same")(pool4)
    convm = layers.Conv2D(start_neurons * 8, (k_sz, k_sz), activation="relu", padding="same")(convm)
    poolm = layers.AveragePooling2D((1, 2))(convm)

    # 3
    two_before_output_layer = layers.Flatten()(poolm)
    one_before_output_layer = layers.Dense(10,activation='relu')(two_before_output_layer)
    output_layer = layers.Dense(1,activation='relu')(one_before_output_layer)
    x_out_DOA = output_layer

    supreg_net = Model(in0, x_out_DOA, name='supreg')
    
    supreg_net.compile(optimizer=k.optimizers.Adam(learning_rate=lr),
              loss=DOA_loss_fn,
              metrics=[tf.keras.losses.MeanSquaredError()])
    
    return supreg_net, x_out_DOA, one_before_output_layer


def get_custom_model_Elevation_est_MCE(in0, input_shape, T, k_sz=3, long_k_sz=101, lr=0.0003):

    start_neurons = 8

    num_of_functions = input_shape[0]

    x = in0
    x = layers.BatchNormalization()(x)

    # 200
    conv1 = layers.Conv2D(T, (num_of_functions, long_k_sz), activation="relu", padding="same")(x)
    conv1 = layers.Conv2D(T, (num_of_functions, long_k_sz), activation="relu", padding="same")(conv1)
    conv1 = layers.Conv2D(start_neurons * 8, (num_of_functions, k_sz), activation="relu", padding="same")(conv1)
    pool1 = layers.AveragePooling2D((1, 2))(conv1)
    pool1 = layers.Dropout(0.5)(pool1)

    # 100
    conv2 = layers.Conv2D(start_neurons * 8, (k_sz, k_sz), activation="relu", padding="same")(pool1)
    conv2 = layers.Conv2D(start_neurons * 8, (k_sz, k_sz), activation="relu", padding="same")(conv2)
    pool2 = layers.AveragePooling2D((1, 2))(conv2)
    pool2 = layers.Dropout(0.5)(pool2)

    # 50
    conv3 = layers.Conv2D(start_neurons * 4, (k_sz, k_sz), activation="relu", padding="same")(pool2)
    conv3 = layers.Conv2D(start_neurons * 4, (k_sz, k_sz), activation="relu", padding="same")(conv3)
    pool3 = layers.AveragePooling2D((1, 2))(conv3)
    pool3 = layers.Dropout(0.5)(pool3)


    # 25
    conv4 = layers.Conv2D(start_neurons * 2, (k_sz, k_sz), activation="relu", padding="same")(pool3)
    conv4 = layers.Conv2D(start_neurons * 2, (k_sz, k_sz), activation="relu", padding="same")(conv4)
    pool4 = layers.AveragePooling2D((1, 2))(conv4)
    pool4 = layers.Dropout(0.5)(pool4)

    # 12
    convm = layers.Conv2D(start_neurons * 8, (k_sz, k_sz), activation="relu", padding="same")(pool4)
    convm = layers.Conv2D(start_neurons * 8, (k_sz, k_sz), activation="relu", padding="same")(convm)
    poolm = layers.AveragePooling2D((1, 2))(convm)

    # 3
    two_before_output_layer = layers.Flatten()(poolm)
    one_before_output_layer = layers.Dense(10,activation='relu')(two_before_output_layer)
    output_layer = layers.Dense(1,activation='relu')(one_before_output_layer)
    x_out_elevation = output_layer

    supreg_net = Model(in0, x_out_elevation, name='supreg')

    supreg_net.compile(optimizer=k.optimizers.Adam(learning_rate=lr),
              loss=Elevation_my_loss_fn,
              metrics=[tf.keras.losses.MeanSquaredError()])
    
    return supreg_net, x_out_elevation, one_before_output_layer


def get_direct_localization_model_weight_spherical(input_shape, L, k_sz=3, long_k_sz=101, lr=0.0001):
    
    in0 = layers.Input(shape=input_shape)
    
    long_k_sz_in = long_k_sz
    
    Range_model, _, one_before_output_layer_Range = get_custom_model_range_est(in0, input_shape, L, k_sz=3, long_k_sz=long_k_sz_in, lr=0.0003)
    DOA_model, _, one_before_output_layer_DOA = get_custom_model_DOA_est_MCE(in0, input_shape, L, k_sz=3, long_k_sz=long_k_sz_in, lr=0.0003)
    Elevation_model, _, one_before_output_layer_Elevation = get_custom_model_Elevation_est_MCE(in0, input_shape, L, k_sz=3, long_k_sz=long_k_sz_in, lr=0.0003)
    
    try:
      Range_model.load_weights('range_model_MLSP2022_weights')
      DOA_model.load_weights('DOA_model_MLSP2022_weights')
      Elevation_model.load_weights('elevation_model_MLSP2022_weights')
    except:
      print('Weights not loaded')
    
    combined1 = layers.Concatenate(axis=-1)([one_before_output_layer_Range, one_before_output_layer_DOA, one_before_output_layer_Elevation])
    location_est = layers.Dense(3,activation='relu')(combined1)
    
    x_out = location_est

    supreg_net = Model(in0, x_out, name='supreg')
    
    supreg_net.compile(optimizer=k.optimizers.Adam(learning_rate=lr),
              loss=direct_localization_loss_fn_spherical,
              metrics=[tf.keras.losses.MeanSquaredError()])
    
    return supreg_net


T = 100
SNR_dB = np.linspace(-10,20,10)
SNR_lin = 10**(SNR_dB/10)
L = 4 # number of sensors
fmax = 1000 # [Hz]

fft_coef = 1/( np.sqrt(T) )
ifft_coef = np.sqrt(T)

x_receiver = np.array([[150 , 50 , -50 , -150]]).T # x coordinate of the sensors [m]
y_receiver = np.array([[-250 , -250 , -250 , -250]]).T # y coordinate of the sensors [m]
z_receiver = np.array([[10 , 15 , 20 , 25]]).T # y coordinate of the sensors [m]
p_array = np.array([x_receiver.T, y_receiver.T, z_receiver.T])
p_array = np.squeeze(p_array)

c = 1500 # [m/sec]
h = 50
w = np.array(range(T))
w = (2*np.pi*fmax*w)/T # frequencies


##########################################
# Import recorded noise
##########################################
# For the real ocean recorded noise, please email me to request the data
#
# Import recorded noise from KAM11
# V_recorded_mat = np.load('noise_KAM11_1kHz_at_13kHz.npy')
# num_of_files, num_of_channels, num_of_samp = V_recorded_mat.shape

##########################################
# Testing data
##########################################
N_test = 10000

p_source_test = np.zeros( (N_test,np.size(SNR_dB),3) )
range_source_test = np.zeros( (N_test,np.size(SNR_dB),1) )
DOA_source_test = np.zeros( (N_test,np.size(SNR_dB),1) )
phi_source_test = np.zeros( (N_test,np.size(SNR_dB),1) )
X_test = np.zeros( (L,T,N_test,np.size(SNR_dB)),dtype = 'complex_')
X_test_input = np.zeros( (N_test,np.size(SNR_dB),L,T,2), dtype = 'complex_' )
X_test_correlation = np.zeros( ( N_test,np.size(SNR_dB),int((L-1)*L*0.5+L),2*T,2),dtype = 'complex_')

surface_time_delay_std_percentage = 0 # 0 - "static" setting, 0.01 (i.e., 1% perturbation) - "dynamic" setting

for snr_idx in tqdm(range(np.size(SNR_dB))):
    noise_power = 1/SNR_lin[snr_idx]
    
    for idx in range(N_test):
        x0 = np.multiply(300,np.random.rand())-150 # true x coordinate of emitters [m]
        y0 = np.multiply(100,np.random.rand())-100 # true y coordinate of emitters [m]
        z0 = 5 + np.multiply(40,np.random.rand()); # the depth of the source
        source_position = np.array( [x0 , y0 , z0] )
        p_source_test[idx,snr_idx,:] = source_position

        range_source_test[idx,snr_idx,0] = np.linalg.norm([x0 , y0 , z0])
        DOA_source_test[idx,snr_idx,0] = ( np.arctan2(y0, x0) + np.pi )
        phi_source_test[idx,snr_idx,0] = ( np.arctan( np.sqrt( y0**2 + x0**2 )/z0 ) )
        
        surface_time_delay_std = (np.linalg.norm(source_position-np.mean(p_array,1))/c)*surface_time_delay_std_percentage 

        S = (np.random.randn(1,T) + np.random.randn(1,T)*1j)/np.sqrt(2)
        for l_ind in range(L):
            b = 0.995+0.1*(np.random.randn(3,1) + np.random.randn(3,1)*1j)/np.sqrt(2) # E[|b_{rl}|^2]=1
            distance_source_reciver = np.sqrt( (x0-x_receiver[l_ind])**2 + (y0-y_receiver[l_ind])**2 + (z0-z_receiver[l_ind])**2 )
            tau1 = distance_source_reciver/c
            distance_surface_bounce = np.sqrt( distance_source_reciver**2 + (z0+z_receiver[l_ind])**2 )
            tau2 = distance_surface_bounce/c
            
            flag_surface_perturbation = 0
            while flag_surface_perturbation==0:
                tau2 = distance_surface_bounce/c + surface_time_delay_std*np.random.randn(1)
                if tau2>0:
                    flag_surface_perturbation = 1
                    
            distance_bottom_bounce = np.sqrt( distance_source_reciver**2 + (2*h - z0 - z_receiver[l_ind])**2 )
            tau3 = distance_bottom_bounce/c
            tau = np.array( [tau1 , tau2 , tau3] )
            
			###########################################################################
            # For the real ocean recorded noise, please email me to request the data
			#
			# real ocean recorded noise (KAM11)
			# file_ind = np.random.randint(2)+(num_of_files-4)
            # channel_ind = np.random.randint(2)+num_of_channels-2
            # start_ind_rec = np.random.randint(num_of_samp-101)
            # V_selected = V_recorded_mat[file_ind,channel_ind,start_ind_rec:start_ind_rec+T]
            # V_selected_norm = V_selected/np.std(V_selected)
            # V = ( np.sqrt(noise_power)*V_selected_norm )/np.sqrt(2)
			###########################################################################
			
			# CN white noise
            V = np.sqrt(noise_power)*(np.random.randn(1,T) + np.random.randn(1,T)*1j)/np.sqrt(2)
            
            D_and_b = np.matmul( b.T, np.exp(-1j*np.outer(tau,w)) )
            x_bar_l = np.multiply(S,D_and_b) + V
            X_test[l_ind,:,idx,snr_idx] = np.multiply(ifft_coef,np.fft.ifft(x_bar_l)) 
            X_test_input[idx,snr_idx,l_ind,:,0] = np.real(X_test[l_ind,:,idx,snr_idx])
            X_test_input[idx,snr_idx,l_ind,:,1] = np.imag(X_test[l_ind,:,idx,snr_idx])

        counter_ind = 0
        for l1_ind in range(L):
            for l2_ind in range(l1_ind,L):
                X_corr_temp = np.correlate(X_test[l1_ind,:,idx,snr_idx],X_test[l2_ind,:,idx,snr_idx],"full")/T
                X_corr_temp = np.append(X_corr_temp,0)
                X_test_correlation[idx,snr_idx,counter_ind,:,0] = X_corr_temp.real
                X_test_correlation[idx,snr_idx,counter_ind,:,1] = X_corr_temp.imag
                counter_ind+=1
##########################################

shape_vector = X_test_correlation.shape
input_shape = shape_vector[2:]
k_sz=3
long_k_sz = T

nn_model = get_direct_localization_model_weight_spherical(input_shape, L, k_sz, long_k_sz)

nn_model.load_weights("direct_localization_model_static")
# to load the weights for the "dynamic" setting, uncomment the next line and comment the line above
# nn_model.load_weights("direct_localization_model_dynamic")

error_test_set = np.zeros( (N_test,3,np.size(SNR_dB)) )
for idx_snr in tqdm(range(np.size(SNR_dB))):
    for idx_test in range(N_test):
        X_test_sample_valid = np.expand_dims(X_test_correlation[idx_test,idx_snr,:], axis=0)
        location_est_test = np.squeeze(nn_model.predict(X_test_sample_valid))
        error_test_set[idx_test,0,idx_snr] = np.mod(location_est_test[0],2*np.pi)-DOA_source_test[idx_test,idx_snr]
        
        error_test_set[idx_test,1,idx_snr] = location_est_test[1]-range_source_test[idx_test,idx_snr]
        
        error_test_set[idx_test,2,idx_snr] = (np.mod(location_est_test[2]+np.pi/2,np.pi)-np.pi/2 )-phi_source_test[idx_test,idx_snr]
        
location_est_test_matrix = np.zeros( (np.size(SNR_dB),N_test,3 ) )
for idx_snr in tqdm(range(np.size(SNR_dB))):
    for idx_test in range(N_test):
        location_est_test_matrix[idx_snr,idx_test,0] = error_test_set[idx_test,0,idx_snr] + DOA_source_test[idx_test,idx_snr]
        location_est_test_matrix[idx_snr,idx_test,1] = error_test_set[idx_test,1,idx_snr] + range_source_test[idx_test,idx_snr]
        location_est_test_matrix[idx_snr,idx_test,2] = error_test_set[idx_test,2,idx_snr] + phi_source_test[idx_test,idx_snr]
        
squared_error_position = np.zeros( (np.size(SNR_dB),N_test ) )
for idx_snr in tqdm(range(np.size(SNR_dB))):
    for idx_test in range(N_test):
        est_DOA = location_est_test_matrix[idx_snr,idx_test,0]
        est_range = location_est_test_matrix[idx_snr,idx_test,1]
        est_elevation = location_est_test_matrix[idx_snr,idx_test,2]
        
        true_DOA = DOA_source_test[idx_test,idx_snr]
        true_range = range_source_test[idx_test,idx_snr]
        true_elevation = phi_source_test[idx_test,idx_snr]
        
        termA = np.sin(true_elevation)*np.sin(est_elevation)*np.cos(true_DOA-est_DOA)
        termB = np.cos(true_elevation)*np.cos(est_elevation)
        
        squared_error_position[idx_snr,idx_test] = est_range**2 + true_range**2 -2*est_range*true_range*( termA + termB )
        
RMSE_position = np.sqrt(np.mean( squared_error_position, axis=1 ))
print(RMSE_position)
sio.savemat('RMSE_position_mat.mat', {'RMSE_position_CNN': RMSE_position})
sio.savemat('squared_error_position.mat', {'squared_error_position_CNN': squared_error_position})


estimation_errors = np.zeros( (np.size(SNR_dB),N_test,3 ) )
for idx_snr in tqdm(range(np.size(SNR_dB))):
    for idx_test in range(N_test):
        est_DOA = location_est_test_matrix[idx_snr,idx_test,0]
        est_range = location_est_test_matrix[idx_snr,idx_test,1]
        est_elevation = location_est_test_matrix[idx_snr,idx_test,2]
        
        true_DOA = DOA_source_test[idx_test,idx_snr]
        true_range = range_source_test[idx_test,idx_snr]
        true_elevation = phi_source_test[idx_test,idx_snr]
        
        estimation_errors[idx_snr,idx_test,0] = est_DOA - true_DOA
        estimation_errors[idx_snr,idx_test,1] = est_range - true_range
        estimation_errors[idx_snr,idx_test,2] = est_elevation - true_elevation
        
sio.savemat('estimation_errors.mat', {'estimation_errors_CNN': estimation_errors})
#------------- END CODE --------------