# Train model for range estimation
# Based on the paper:
#			 "Direct Localization in Underwater Acoustics via Convolutional 
#             Neural Networks: A Data-Driven Approach", Amir Weiss,
#             Toros Arikan, and Gregory W. Wornell,
#			  in Proc. of MLSP, August 2022.
#
# See also: supplementary_materials_DLOC_MLSP2022.pdf for more information about this Python package
# Author: Amir Weiss
# email: amirweiss15@gmail.com
# Website: http://www.weissamir.com
# July 2022; Last revision: 11-Jul-2022
#------------- BEGIN CODE --------------
import tensorflow as tf
import tensorflow.keras as k
from tensorflow.keras import layers
from tensorflow.keras.models import Model

import numpy as np
import matplotlib.pyplot as plt
from tqdm import tqdm
import time
import pickle
import os

import random 

import range_model_MSE as range_est
import scipy.io as sio

T = 100
SNR_dB = np.linspace(-10,30,20)
SNR_lin = 10**(SNR_dB/10)
L = 4 # number of sensors
fmax = 1000 # [Hz]

fft_coef = 1/( np.sqrt(T) )
ifft_coef = np.sqrt(T)

x_receiver = np.array([[150 , 50 , -50 , -150]]).T # x coordinate of the sensors [m]
y_receiver = np.array([[-250 , -250 , -250 , -250]]).T # y coordinate of the sensors [m]
z_receiver = np.array([[10 , 15 , 20 , 25]]).T # y coordinate of the sensors [m]
p_array = np.array([x_receiver.T, y_receiver.T, z_receiver.T])
p_array = np.squeeze(p_array)

c = 1500 # [m/sec]
h = 50
w = np.array(range(T))
w = (2*np.pi*fmax*w)/T # frequencies

surface_time_delay_std_percentage = 0 # 0 - "static" setting, 0.01 (i.e., 1% perturbation) - "dynamic" setting

##########################################
# Import recorded noise
##########################################
# For the real ocean recorded noise, please email me to request the data
#
# Import recorded noise from KAM11
# V_recorded_mat = np.load('noise_KAM11_1kHz_at_13kHz.npy')
# num_of_files, num_of_channels, num_of_samp = V_recorded_mat.shape

##########################################
# Training data
##########################################
N_train = 200000//SNR_dB.size

p_source_train, range_source_train = [], []
X_train = np.zeros( (L,T,N_train,np.size(SNR_dB)),dtype = 'complex_')

X_train_input = np.zeros( (N_train,np.size(SNR_dB),L,T,2), dtype = 'complex_' )
X_train_correlation = np.zeros( ( N_train*np.size(SNR_dB),int((L-1)*L*0.5+L),2*T,2),dtype = 'complex_')

for snr_idx in tqdm(range(np.size(SNR_dB))):
    noise_power = 1/SNR_lin[snr_idx]
    
    for idx in range(N_train):
        x0 = np.multiply(300,np.random.rand())-150 # true x coordinate of emitters [m]
        y0 = np.multiply(100,np.random.rand())-100 # true y coordinate of emitters [m]
        z0 = 5 + np.multiply(40,np.random.rand()); # the depth of the source
        source_position = np.array( [x0 , y0 , z0] )
        p_source_train.append(source_position)
        range_source_train.append( np.linalg.norm([x0 , y0 , z0]) )
        
        surface_time_delay_std = (np.linalg.norm(source_position-np.mean(p_array,1))/c)*surface_time_delay_std_percentage

        S = (np.random.randn(1,T) + np.random.randn(1,T)*1j)/np.sqrt(2)
        for l_ind in range(L):
            b = 0.995+0.1*(np.random.randn(3,1) + np.random.randn(3,1)*1j)/np.sqrt(2) # E[|b_{rl}|^2]=1
            distance_source_reciver = np.sqrt( (x0-x_receiver[l_ind])**2 + (y0-y_receiver[l_ind])**2 + (z0-z_receiver[l_ind])**2 )
            tau1 = distance_source_reciver/c
            distance_surface_bounce = np.sqrt( distance_source_reciver**2 + (z0+z_receiver[l_ind])**2 )
            tau2 = distance_surface_bounce/c
            
            flag_surface_perturbation = 0
            while flag_surface_perturbation==0:
                tau2 = distance_surface_bounce/c + surface_time_delay_std*np.random.randn(1)
                if tau2>0:
                    flag_surface_perturbation = 1
            
            distance_bottom_bounce = np.sqrt( distance_source_reciver**2 + (2*h - z0 - z_receiver[l_ind])**2 )
            tau3 = distance_bottom_bounce/c
            tau = np.array( [tau1 , tau2 , tau3] )
            
            ###########################################################################
            # For the real ocean recorded noise, please email me to request the data
            #
            # real ocean recorded noise (KAM11)
#             file_ind = np.random.randint(4)+(num_of_files-6)
#             channel_ind = np.random.randint(2)+num_of_channels-4
#             start_ind_rec = np.random.randint(num_of_samp-101)
#             V_selected = V_recorded_mat[file_ind,channel_ind,start_ind_rec:start_ind_rec+T]
#             V_selected_norm = V_selected/np.std(V_selected)
#             V = ( np.sqrt(noise_power)*V_selected_norm )/np.sqrt(2)
            ###########################################################################
            
            # CN white noise
            V = np.sqrt(noise_power)*(np.random.randn(1,T) + np.random.randn(1,T)*1j)/np.sqrt(2)
            
            D_and_b = np.matmul( b.T, np.exp(-1j*np.outer(tau,w)) )
            x_bar_l = np.multiply(S,D_and_b) + V
            X_train[l_ind,:,idx,snr_idx] = np.multiply(ifft_coef,np.fft.ifft(x_bar_l)) 
            X_train_input[idx,snr_idx,l_ind,:,0] = np.real(X_train[l_ind,:,idx,snr_idx])
            X_train_input[idx,snr_idx,l_ind,:,1] = np.imag(X_train[l_ind,:,idx,snr_idx])

        counter_ind = 0
        for l1_ind in range(L):
            for l2_ind in range(l1_ind,L):
                X_corr_temp = np.correlate(X_train[l1_ind,:,idx,snr_idx],X_train[l2_ind,:,idx,snr_idx],"full")/T
                X_corr_temp = np.append(X_corr_temp,0)
                X_train_correlation[snr_idx*N_train+idx,counter_ind,:,0] = X_corr_temp.real
                X_train_correlation[snr_idx*N_train+idx,counter_ind,:,1] = X_corr_temp.imag
                counter_ind+=1

p_source_train = np.array(p_source_train)
range_source_train = np.array(range_source_train)

##########################################
# Validation data
##########################################
N_val = 5000

p_source_val, range_source_val = [], []
X_val = np.zeros( (L,T,N_val,np.size(SNR_dB)),dtype = 'complex_')
X_val_input = np.zeros( (N_val,np.size(SNR_dB),L,T,2), dtype = 'complex_' )
X_val_correlation = np.zeros( ( N_val*np.size(SNR_dB),int((L-1)*L*0.5+L),2*T,2),dtype = 'complex_')

for snr_idx in tqdm(range(np.size(SNR_dB))):
    noise_power = 1/SNR_lin[snr_idx]
    
    for idx in range(N_val):
        x0 = np.multiply(300,np.random.rand())-150 # true x coordinate of emitters [m]
        y0 = np.multiply(100,np.random.rand())-100 # true y coordinate of emitters [m]
        z0 = 5 + np.multiply(40,np.random.rand()); # the depth of the source
        source_position = np.array( [x0 , y0 , z0] )
        p_source_val.append(source_position)

        range_source_val.append( np.linalg.norm([x0 , y0 , z0]) )

        surface_time_delay_std = (np.linalg.norm(source_position-np.mean(p_array,1))/c)*surface_time_delay_std_percentage

        S = (np.random.randn(1,T) + np.random.randn(1,T)*1j)/np.sqrt(2)
        for l_ind in range(L):
            b = 0.995+0.1*(np.random.randn(3,1) + np.random.randn(3,1)*1j)/np.sqrt(2) # E[|b_{rl}|^2]=1
            distance_source_reciver = np.sqrt( (x0-x_receiver[l_ind])**2 + (y0-y_receiver[l_ind])**2 + (z0-z_receiver[l_ind])**2 )
            tau1 = distance_source_reciver/c
            distance_surface_bounce = np.sqrt( distance_source_reciver**2 + (z0+z_receiver[l_ind])**2 )
            tau2 = distance_surface_bounce/c
            
            flag_surface_perturbation = 0
            while flag_surface_perturbation==0:
                tau2 = distance_surface_bounce/c + surface_time_delay_std*np.random.randn(1)
                if tau2>0:
                    flag_surface_perturbation = 1
                    
            distance_bottom_bounce = np.sqrt( distance_source_reciver**2 + (2*h - z0 - z_receiver[l_ind])**2 )
            tau3 = distance_bottom_bounce/c
            tau = np.array( [tau1 , tau2 , tau3] )
            
            ###########################################################################
            # For the real ocean recorded noise, please email me to request the data
            #
            # real ocean recorded noise (KAM11)
#             file_ind = np.random.randint(4)+(num_of_files-6)
#             channel_ind = np.random.randint(2)+num_of_channels-4
#             start_ind_rec = np.random.randint(num_of_samp-101)
#             V_selected = V_recorded_mat[file_ind,channel_ind,start_ind_rec:start_ind_rec+T]
#             V_selected_norm = V_selected/np.std(V_selected)
#             V = ( np.sqrt(noise_power)*V_selected_norm )/np.sqrt(2)
            ###########################################################################
            
            # CN white noise
            V = np.sqrt(noise_power)*(np.random.randn(1,T) + np.random.randn(1,T)*1j)/np.sqrt(2)
            
            D_and_b = np.matmul( b.T, np.exp(-1j*np.outer(tau,w)) )
            x_bar_l = np.multiply(S,D_and_b) + V
            X_val[l_ind,:,idx,snr_idx] = np.multiply(ifft_coef,np.fft.ifft(x_bar_l)) 
            X_val_input[idx,snr_idx,l_ind,:,0] = np.real(X_val[l_ind,:,idx,snr_idx])
            X_val_input[idx,snr_idx,l_ind,:,1] = np.imag(X_val[l_ind,:,idx,snr_idx])

        counter_ind = 0
        for l1_ind in range(L):
            for l2_ind in range(l1_ind,L):
                X_corr_temp = np.correlate(X_val[l1_ind,:,idx,snr_idx],X_val[l2_ind,:,idx,snr_idx],"full")/T
                X_corr_temp = np.append(X_corr_temp,0)
                X_val_correlation[snr_idx*N_val+idx,counter_ind,:,0] = X_corr_temp.real
                X_val_correlation[snr_idx*N_val+idx,counter_ind,:,1] = X_corr_temp.imag
                counter_ind+=1

p_source_val = np.array(p_source_val)
range_source_val = np.array(range_source_val)
##########################################

num_of_epochs = 300

shape_vector = X_train_correlation.shape
input_shape = shape_vector[1:]
k_sz=3
long_k_sz = T

checkpoint_filepath = './tmp_range_model_MLSP2022/checkpoint'

model_checkpoint_callback = tf.keras.callbacks.ModelCheckpoint(
    filepath=checkpoint_filepath,
    save_weights_only=True,
    monitor='val_loss',
    mode='min',
    save_best_only=True)


nn_model = range_est.get_custom_model_range_est(input_shape, L, k_sz, long_k_sz)
nn_model.fit(X_train_correlation, range_source_train, validation_data=(X_val_correlation,range_source_val), epochs=num_of_epochs, batch_size=30, shuffle=True, verbose=1, callbacks=[model_checkpoint_callback])

nn_model.save("range_model_MLSP2022")
nn_model.save_weights("range_model_MLSP2022_weights")
#------------- END CODE --------------